/*
 * MultipleAlignment.h
 *
 *  Created on: Aug 12, 2009
 *      Author: Oleksiy Kohany <kohany@girinst.org>
 */

#ifndef MULTIPLEALIGNMENT_H_
#define MULTIPLEALIGNMENT_H_

#include <vector>
#include <iterator>

#include "ScoreMatrix.h"
#include "BioSequence.h"
#include "BioAlignment.h"

#ifndef _NO_ALN
#include "ThreadedAligner.h"
#endif

#ifdef _USE_GSL_
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#else
#include <boost/math/distributions/binomial.hpp>
#endif

using namespace std;

typedef vector<BioSubSequence> SEQ_VECTOR;
typedef list<BioAlignment> ALN_LIST;

enum TAXON {
   UNKNOWN, MAM, VRT, PLN
};

class MultipleAlignment {
   SEQ_VECTOR alnvec;
   double score;
   static bool DEBUG;
#ifndef _NO_ALN
   static ThreadedAligner aligner;
#endif
public:
   typedef SEQ_VECTOR::iterator iterator;
   typedef SEQ_VECTOR::const_iterator const_iterator;
   typedef SEQ_VECTOR::reverse_iterator reverse_iterator;
   typedef SEQ_VECTOR::const_reverse_iterator const_reverse_iterator;
   typedef SEQ_VECTOR::reference reference;
   typedef SEQ_VECTOR::const_reference const_reference;

   MultipleAlignment();
   MultipleAlignment(const MultipleAlignment& other);
   virtual ~MultipleAlignment();
   MultipleAlignment& operator =(const MultipleAlignment& other);

   iterator begin() {
      return alnvec.begin();
   }
   iterator end() {
      return alnvec.end();
   }
   const_iterator begin() const {
      return alnvec.begin();
   }
   const_iterator end() const {
      return alnvec.end();
   }
   reverse_iterator rbegin() {
      return alnvec.rbegin();
   }
   reverse_iterator rend() {
      return alnvec.rend();
   }
   const_reverse_iterator rbegin() const {
      return alnvec.rbegin();
   }
   const_reverse_iterator rend() const {
      return alnvec.rend();
   }
   iterator erase(iterator position) {
      return alnvec.erase(position);
   }
   iterator erase(iterator first, iterator last) {
      return alnvec.erase(first, last);
   }
   void reserve(size_t n) {
      alnvec.reserve(n);
   }
   void push_back(const BioSubSequence& x) {
      alnvec.push_back(x);
   }
   reference front() {
      return alnvec.front();
   }
   const_reference front() const {
      return alnvec.front();
   }
   reference back() {
      return alnvec.back();
   }
   const_reference back() const {
      return alnvec.back();
   }
   size_t size() const {
      return alnvec.size();
   }
   reference operator[](size_t n) {
      return alnvec[n];
   }
   const_reference operator[](size_t n) const {
      return alnvec[n];
   }
   void clear();

#ifndef _NO_ALN
   static void setGapIni(int gapIni);
   static void setGapExt(int gapExt);
   static void setAlignmentMatrix(const ScoreMatrix& mat);
   static void setAlignmentMethod(ALNALG method);
#endif
   void subAlignment(MultipleAlignment& subaln, size_t pos, size_t length) const;
   double calcAlignmentScores(const string& ref, vector<double>& columnScores, vector<double>& rowScores, const ScoreMatrix* matrix = NULL) const;
   double calcAlignmentScores(vector<double>& columnScores, vector<double>& rowScores, const ScoreMatrix* matrix = NULL) const;
   double getScore() const;
   int read(istream& is, bool trimGaps = false);
   int print(ostream& os);

   void restoreCpG(const BioSequence& icons, BioSequence& ocons, TAXON species) const;
   template<class ForwardIterator>
   void restoreCpG(const BioSequence& icons, BioSequence& ocons, ForwardIterator begin, ForwardIterator end, TAXON species) const;

   void autoConsensus(BioSequence& cons, const ScoreMatrix* conmatptr, int minNonGapCount = 1, TAXON species = UNKNOWN,
         bool finalize = false, bool keep_gaps = false) const;
   void getConsensus(string& cons, const ScoreMatrix* matptr = NULL, int minNonGapCount = 1) const;
   template<class ForwardIterator>
   static void getConsensus(string& sCON, ForwardIterator beg, ForwardIterator end, const ScoreMatrix* matptr = NULL, int minNonGapCount = 1);
#ifndef _NO_ALN
   template<class ForwardIterator>
   static void autoConsensus(RatedSequence& sCON, const typename std::iterator_traits<ForwardIterator>::reference ref,
         ForwardIterator beg, ForwardIterator end, const ScoreMatrix* conmatptr, double& outscore, int minNonGapCount = 1,
         TAXON species = UNKNOWN, bool withRefSeq = true, bool finalize = false);

   void correctLowScoringRegions(BioSubSequence& sCON, const ScoreMatrix* matptr = NULL) const;
   template<class ForwardIterator, class OutputIterator>
   double align(const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg, ForwardIterator end, OutputIterator out);
   template<class ForwardIterator>
   double construct(const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg, ForwardIterator end);
   template<class ForwardIterator>
   ForwardIterator autoAlign(ForwardIterator beg, ForwardIterator end);
#endif
};

void adjustReference(string& ref, const string& seq, int offset);
void adjustAlignment(const string& ref1, string& ref2, string& seq, int offset);

double inline pts_mam(double sim);
double inline pCA_TG(double pTs, double Sc);
double inline pTG_CA(double pTs, double Sc);
double inline pCA_TA(double pTs, double Sc);
double inline pTG_TA(double pTs, double Sc);
double inline pCA_CA(double pTs, double Sc);
double inline pTG_TG(double pTs, double Sc);
double inline pTG_CG(double pTs, double Sc);
double inline pCA_CG(double pTs, double Sc);
double inline pTA_YR(double pTs, double Sc);
double inline pTG_CR(double pTs, double Sc);
double inline pTG_TR(double pTs, double Sc);
double inline pCA_TR(double pTs, double Sc);
double inline pCA_YA(double pTs, double Sc);
double inline pTG_YA(double pTs, double Sc);
double inline pCA_YG(double pTs, double Sc);
double inline pCA_CR(double pTs, double Sc);
double inline pTG_YG(double pTs, double Sc);
double inline pCA_YR(double pTs, double Sc);
double inline pTG_YR(double pTs, double Sc);

template<class ForwardIterator>
void MultipleAlignment::getConsensus(string& sCON, ForwardIterator beg, ForwardIterator end, const ScoreMatrix* matptr, int minNonGapCount) {
   ScoreMatrix defaultMatrix;
   defaultMatrix.loadMatrixFromString(ScoreMatrix::CONMATRIX);
   const ScoreMatrix& matrix = matptr == NULL ? defaultMatrix : *matptr;

   string alphabeth = matrix.getAlphabeth();

   alphabeth.push_back(ScoreMatrix::GAPCHAR);
   size_t alnWidth = (*beg).size();
   size_t alnHeight = distance(beg, end);
   sCON.assign(alnWidth, ScoreMatrix::GAPCHAR);
   double scores[128];
   size_t conStart = -1;
   size_t conEnd = -1;
   for (int x = alnWidth - 1; x >= 0; x--) {
      double maxScore = -1000000;
      char maxChar = ScoreMatrix::GAPCHAR;
      for (int i = alphabeth.size() - 1; i >= 0; i--) {
         double score = 0;
         int count = 0;
         for (ForwardIterator sptr = beg; sptr != end; sptr++) {
            if ((*sptr)[x] != ScoreMatrix::GAPCHAR && (*sptr)[x] != ScoreMatrix::RPADDING && (*sptr)[x] != ScoreMatrix::LPADDING) {
               count++;
            }
            try {
               score += matrix.getScore(alphabeth[i], (*sptr)[x]);
            } catch (logic_error exc) {
               //cerr << (*sptr).getName() << endl;
               cerr << exc.what() << endl;
            } catch (runtime_error exc) {
               cerr << exc.what() << endl;
            }
         }
         if (score > maxScore) {
            if (count >= minNonGapCount) {
               maxScore = score;
               maxChar = alphabeth[i];
            }
         }
      }
      sCON[x] = maxChar;
   }
}

/**
 * First sequence in the alignment vector is assumed to be consensus
 * that requires restoration of CpG doublets.
 * species is needed to estimate ratio of types of mutations: transitions to transversions.
 */
template<class ForwardIterator>
void MultipleAlignment::restoreCpG(const BioSequence& icons, BioSequence& ocons, ForwardIterator begin, ForwardIterator end, TAXON species) const {
   double sim = 0;
   vector<string> masks;
   masks.push_back("CG");
   int matchCnt = 0;
   int missCnt = 0;
   int malnSize = 0;

   if (&ocons != &icons) {
      ocons.assign(icons);
   }
   ocons.toupper();
   /**
    * Copy input alignment removing columns corresponding to consensus gaps
    */
   for (ForwardIterator ptr = begin; ptr != end; ptr++) {
      AlnStat stat = BioAlignment::getStat(ocons, *ptr, &masks);
      matchCnt += stat.iMCnt;
      missCnt += stat.iMMCnt;
      malnSize++;
   }

   sim = (double) matchCnt / (matchCnt + missCnt);
   if (DEBUG)
      cerr << "SIM=" << sim << endl;

   double pTs = 0;
   switch (species) {
   case UNKNOWN:
   case VRT:
   case PLN:
   case MAM:
      pTs = pts_mam(sim); // This implementation of probability of transition is biologically correct for mammals only.
      // More research is needed for other types of organisms.
   }
   string::iterator sptr1, sptr2;
   sptr1 = ocons.begin();

   for (sptr1 = ocons.begin(); sptr1 != ocons.end(); sptr1++) {
      double pTG = 0;
      double pCA = 0;
      double pTA = 0;
      while (!isalpha(*sptr1) && sptr1 != ocons.end()) {
         sptr1++;
      }
      if (sptr1 == ocons.end()) {
         break;
      }
      sptr2 = sptr1 + 1;
      while (!isalpha(*sptr2) && sptr2 != ocons.end()) {
         sptr2++;
      }
      if (sptr2 == ocons.end()) {
         break;
      }

      int i1 = sptr1 - ocons.begin();
      int i2 = sptr2 - ocons.begin();

      string doublet;
      doublet.clear();
      doublet.push_back(*sptr1);
      doublet.push_back(*sptr2);
      switch (*sptr1) {
      case 'T':
         switch (*sptr2) {
         case 'G':
            pCA = pCA_TG(pTs, sim);
            break;
         case 'A':
            pTG = pCA = pCA_TA(pTs, sim);
            break;
         case 'R':
            pCA = pCA_TR(pTs, sim);
            break;
         default:
            continue;
         }
         break;
      case 'C':
         switch (*sptr2) {
         case 'A':
            pTG = pTG_CA(pTs, sim);
            break;
         case 'R':
            pTG = pTG_CR(pTs, sim);
            break;
         default:
            continue;
         }
         break;
      case 'Y':
         switch (*sptr2) {
         case 'G':
            pCA = pCA_YG(pTs, sim);
            break;
         case 'A':
            pTG = pTG_YA(pTs, sim);
            break;
         case 'R':
            pTA = pTA_YR(pTs, sim);
            break;
         default:
            continue;
         }
         break;
      default:
         continue;
      }

      int nTG = 0;
      int nCA = 0;
      int nTA = 0;
      int nCG = 0;
      int nCON = 0;
      int nDel = 0;
      for (ForwardIterator lptr = begin; lptr != end; lptr++) {
         if ((*lptr)[i1] == *sptr1 && (*lptr)[i2] == *sptr2) {
            nCON++;
         }
         if ((*lptr)[i1] == 'T' && (*lptr)[i2] == 'G') {
            nTG++;
            continue;
         }
         if ((*lptr)[i1] == 'C' && (*lptr)[i2] == 'A') {
            nCA++;
            continue;
         }
         if ((*lptr)[i1] == 'T' && (*lptr)[i2] == 'A') {
            nTA++;
            continue;
         }
         if ((*lptr)[i1] == 'C' && (*lptr)[i2] == 'G') {
            nCG++;
            continue;
         }
         if ((*lptr)[i1] == ScoreMatrix::GAPCHAR || (*lptr)[i2] == ScoreMatrix::GAPCHAR || (*lptr)[i1] == ScoreMatrix::LPADDING || (*lptr)[i1] == ScoreMatrix::RPADDING || (*lptr)[i2]
               == ScoreMatrix::LPADDING || (*lptr)[i2] == ScoreMatrix::RPADDING) {
            nDel++;
         }
      }
      if (DEBUG) {
         cerr << "pos=" << i1 + 1 << "+" << i2 + 1 << " (" << *sptr1 << *sptr2 << ")" << endl;
         cerr << "Sc=" << sim << "; pTs=" << pTs << endl;
         cerr << setw(3) << "nTG=" << nTG << "\t nCA=" << nCA << "\t nTA=" << nTA << "\t nCG=" << nCG << "\t nCON=" << nCON << "\tnDel=" << nDel << endl;
         cerr << "pTG=" << pTG << endl;
         cerr << "pCA=" << pCA << endl;
         cerr << "pTA=" << pTA << endl;
      }
      double p;
      if (pTG > 0.0) {
         if (doublet == "CR" || doublet == "YA") {
#ifdef _USE_GSL_
            p = gsl_cdf_binomial_P(nTG, pTG, malnSize - nDel);
#else
            p = boost::math::cdf(boost::math::binomial(malnSize - nDel, pTG), nTG);
#endif
            if (DEBUG)
               cerr << "pTG(" << nTG << "-, " << malnSize - nDel << ")=" << p << endl;
            if (p < 0.05) {
               if (DEBUG)
                  cerr << "=====================" << endl;
               continue;
            }
         } else {
#ifdef _USE_GSL_
            p = gsl_cdf_binomial_Q(nTG, pTG, malnSize - nDel) + gsl_ran_binomial_pdf(nTG, pTG, malnSize - nDel);
#else
            p = boost::math::cdf(boost::math::complement(boost::math::binomial(malnSize - nDel, pTG), nTG)) + boost::math::pdf(boost::math::binomial(malnSize - nDel, pTG), nTG);
#endif
            if (DEBUG)
               cerr << "qTG(" << nTG << "+, " << malnSize - nDel << ")=" << p << endl;
            if (p > 0.05) {
               if (DEBUG)
                  cerr << "=====================" << endl;
               continue;
            }
         }
      }
      if (pCA > 0.0) {
         if (doublet == "TR" || doublet == "YG") {
#ifdef _USE_GSL_
            p = gsl_cdf_binomial_P(nCA, pCA, malnSize - nDel);
#else
            p = boost::math::cdf(boost::math::binomial(malnSize - nDel, pCA), nCA);
#endif
            if (DEBUG)
               cerr << "pCA(" << nCA << "-, " << malnSize - nDel << ")=" << p << endl;
            if (p < 0.05) {
               if (DEBUG)
                  cerr << "=====================" << endl;
               continue;
            }
         } else {
#ifdef _USE_GSL_
            p = gsl_cdf_binomial_Q(nCA, pCA, malnSize - nDel) + gsl_ran_binomial_pdf(nCA, pCA, malnSize - nDel);
#else
            p = boost::math::cdf(boost::math::complement(boost::math::binomial(malnSize - nDel, pCA), nCA)) + boost::math::pdf(boost::math::binomial(malnSize - nDel, pCA), nCA);
#endif
            if (DEBUG)
               cerr << "qCA(" << nCA << "+, " << malnSize - nDel << ")=" << p << endl;
            if (p > 0.05) {
               if (DEBUG)
                  cerr << "=====================" << endl;
               continue;
            }
         }
      }
      if (pTA > 0.0) {
#ifdef _USE_GSL_
         p = gsl_cdf_binomial_Q(nTA, pTA, malnSize - nDel) + gsl_ran_binomial_pdf(nTA, pTA, malnSize - nDel);
#else
         p = boost::math::cdf(boost::math::complement(boost::math::binomial(malnSize - nDel, pTA), nTA)) + boost::math::pdf(boost::math::binomial(malnSize - nDel, pTA), nTA);
#endif
         if (DEBUG)
            cerr << "qTA(" << nTA << "+, " << malnSize - nDel << ")=" << p << endl;
         if (p <= 0.05) {
            if (DEBUG)
               cerr << "=====================" << endl;
            continue;
         }
      }
      cerr << "Restored pos " << i1 + 1 << "+" << i2 + 1 << " " << *sptr1 << *sptr2 << "=>CG" << endl;
      if (DEBUG)
         cerr << "=====================" << endl;
      *sptr1 = 'C';
      *sptr2 = 'G';
      sptr1++;
   }
}

#ifndef _NO_ALN

template<class ForwardIterator>
void MultipleAlignment::autoConsensus(RatedSequence& sCON, const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg, ForwardIterator end, const ScoreMatrix* conmatptr, double& outscore, int minNonGapCount,
      TAXON species, bool withRefSeq, bool finalize) {
   MultipleAlignment maln;

   outscore = maln.construct(ref, beg, end);

   if (outscore > 0) {
      if (!withRefSeq) {
         maln.erase(maln.begin());
      }
      maln.autoConsensus(sCON, conmatptr, minNonGapCount, species, finalize);
      sCON.setScore(outscore);
      sCON.setName(ref.getName());
   }
}
template<class ForwardIterator, class OutputIterator>
double MultipleAlignment::align(const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg, ForwardIterator end, OutputIterator out) {
   double score = 0;
   double sumScore = 0;
   BioSubSequence sREF(ref);
   sREF.setFrom(1);
   sREF.setTo(sREF.size());
   BioAlignment aln;
   ForwardIterator slsptr;
   ALN_LIST alnls;

   sumScore = aligner.align(ref, beg, end, back_insert_iterator<ALN_LIST> (alnls));

   //double minScore = 0.3 * sumScore / alnls.size();
   double minScore = 0;
   ALN_LIST::iterator lptr = alnls.begin();
   int removedCount = 0;
   int malnSize = 0;
   while (lptr != alnls.end()) {
      if ((*lptr).getScore() >= minScore) {
         adjustReference(sREF, (*lptr).getCon(), (*lptr).getCon().getFrom() - 1);
         lptr++;
         malnSize++;
      } else {
         lptr = alnls.erase(lptr);
         removedCount++;
      }
   }
   if (removedCount) {
      //cerr << "Alignments removed/remaining : " << removedCount << " / " << malnSize << " ; with score threshold = " << minScore << endl;
   }
   *out = sREF;
   ++out;
   for (lptr = alnls.begin(); lptr != alnls.end(); lptr++) {
      adjustAlignment(sREF, (*lptr).getCon(), (*lptr).getSeq(), (*lptr).getCon().getFrom() - 1);
      *out = (*lptr).getSeq();
      ++out;
   }
   return sumScore;
}

template<class ForwardIterator>
double MultipleAlignment::construct(const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg, ForwardIterator end) {
   clear();
   reserve(distance(end, beg) + 1);
   score = align(ref, beg, end, back_insert_iterator<MultipleAlignment> (*this));
   return score;
}

template<class ForwardIterator>
ForwardIterator MultipleAlignment::autoAlign(ForwardIterator beg, ForwardIterator end) {
   double maxScore = 0;
   double score = 0;
   ForwardIterator refPtr = end;
   MultipleAlignment maln;
   for (ForwardIterator ptr = beg; ptr != end; ptr++) {
      try {
         if ((*ptr).size() > 0) {
            score = maln.construct(*ptr, beg, end);
            if (score > maxScore) {
               maxScore = score;
               refPtr = ptr;
               *this = maln;
            }
         }
      } catch (...) {
         cerr << "Alignment with reference " << (*ptr).getName() << " failed" << endl;
      }
   }
   return refPtr;
}

#endif /* _NO_ALN */

#endif /* MULTIPLEALIGNMENT_H_ */
