/******************************************************************
 Copyright 2006 by Michael Farrar.  All rights reserved.
 This program may not be sold or incorporated into a commercial product,
 in whole or in part, without written consent of Michael Farrar.  For
 further information regarding permission for use or reproduction, please
 contact: Michael Farrar at farrar.michael@gmail.com.
 *******************************************************************/

/*
 Written by Michael Farrar, 2006.
 Modified by Oleksiy Kokhanyy (kohany@gmail.com), 2009
 Please send bug reports and/or suggestions to farrar.michael@gmail.com.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "swsse2.h"
#include "score_matrix.h"

#define BUF_SIZE 512

static char CUSTOM_ALPHABETH[256];
static short CUSTOM_ALPHABETH_SIZE = 0;
static short CUSTOM_ALPHABETH_VALUE[256];

char *skipSpaces(char *line)
{
   while (isspace(*line))
   {
      ++line;
   }

   return line;
}

void printMatrix(const signed char* matrix)
{
   int y, x;
   fprintf(stderr, "GAP %d %d\n", gapInit, gapExt);
   fprintf(stderr, "%4c", ' ');
   for (x = 0; x < ALPHA_SIZE; x++)
   {
      fprintf(stderr, "%4c", ALPHABETH[x]);
   }
   fprintf(stderr, "\n");
   for (y = 0; y < ALPHA_SIZE; y++)
   {
      const signed char* row = matrix + (y * ALPHA_SIZE);
      fprintf(stderr, "%4c", ALPHABETH[y]);
      for (x = 0; x < ALPHA_SIZE; x++)
      {
         fprintf(stderr, "%4d", row[x]);
      }
      fprintf(stderr, "\n");
   }
}
/**
 * Reads blast-style and cross_match-style matrices.
 */
signed char *readMatrix(const char *file)
{
   fprintf(stderr, "Loading matrix %s at %s:%d\n", file, __FILE__, __LINE__);
   FILE *fp;
   char line[BUF_SIZE];

   signed char *matrix;

   int done;
   int i;

   int errors = 0;
   int count = 0;

   if ((fp = fopen(file, "r")) == NULL)
   {
      fprintf(stderr, "Unable to open file %s\n", file);
      exit(-1);
   }

   for (i = 0; i < 256; ++i)
   {
      CUSTOM_ALPHABETH_VALUE[i] = -1;
   }

   /* read the first line of the matrix giving the amino acid order */
   done = 0;
   /*
    * Added to store column header values to reuse as row header
    * if we are we have to read cross_match-style matrix.
    */
   while (!done && fgets(line, BUF_SIZE, fp) != NULL)
   {
      char *ptr = skipSpaces(line);
      if (strstr(ptr, "FREQS ") == ptr)
      {
         continue;
      }
      if (*ptr && *ptr != '#')
      {
         /** to read cross_match matrices properly */
         if (strstr(ptr, "GAP ") == ptr)
         {
            int gi, gx;
            if (sscanf(ptr + 3, "%d %d", &gi, &gx) == 2)
            {
               gapInit = gi;
               gapExt = gx;
            }
            continue;
         }
         while (*ptr && *ptr != '#')
         {
            if(CUSTOM_ALPHABETH_VALUE[*ptr] >= 0) {
               fprintf(stderr, "%s, %d: Symbol %c defined twice.\n", __FILE__, __LINE__, *ptr);
               errors++;
            }
            CUSTOM_ALPHABETH_VALUE[tolower(*ptr)] = CUSTOM_ALPHABETH_SIZE;
            CUSTOM_ALPHABETH_VALUE[toupper(*ptr)] = CUSTOM_ALPHABETH_SIZE;
            CUSTOM_ALPHABETH[CUSTOM_ALPHABETH_SIZE++] = toupper(*ptr);
            ptr = skipSpaces(ptr + 1);
         }
         done = 1;
      }
   }

   matrix = (signed char *) malloc(CUSTOM_ALPHABETH_SIZE * CUSTOM_ALPHABETH_SIZE);
   assert(matrix);

   if (errors > 0)
   {
      fprintf(stderr, "Terminating due to errors in matrix's header\n");
      exit(-1);
   }

   int colHeaderIdx = 0;
   /* read the scores for the amino acids */
   while (fgets(line, BUF_SIZE, fp) != NULL)
   {
      signed char *row;
      char *ptr = skipSpaces(line);
      if (*ptr && *ptr != '#')
      {
         char aminoAcid = *ptr;
         /**
          * If it's a cross_match matrix, rows don't have headers.
          * Use previously read column header instead.
          */
         if (isdigit(*ptr) || *ptr == '-' || *ptr == '+')
         {
            aminoAcid = CUSTOM_ALPHABETH[colHeaderIdx++];
            ptr = line - 1;
         }
         int inx = CUSTOM_ALPHABETH_VALUE[aminoAcid];
         if (inx == -1)
         {
            fprintf(stderr, "%s, %d: Unknown amino acid %c in matrix row header\n", __FILE__, __LINE__, aminoAcid);
            continue;
            ++errors;
         }

         row = &matrix[inx * CUSTOM_ALPHABETH_SIZE];

         for (i = 0; i < CUSTOM_ALPHABETH_SIZE; ++i)
         {
            int sign = 1;
            int num = 0;

            ptr = skipSpaces(ptr + 1);

            /* check the sign */
            if (*ptr == '-')
            {
               sign = -1;
               ++ptr;
            }

            do
            {
               if (*ptr >= '0' && *ptr <= '9')
               {
                  num = num * 10 + (*ptr - '0');
                  ptr++;
               }
               else
               {
                  char name[16];
                  char *pName;
                  if (isspace(*ptr))
                  {
                     pName = "space";
                  }
                  else {
                     if (*ptr == 0)
                     {
                        pName = "end of line";
                     }
                     else
                     {
                        name[0] = *ptr;
                        name[1] = 0;
                        pName = name;
                     }
                  }
                  fprintf(stderr, "Row %c Expecting digit found %s\n", aminoAcid, pName);
                  ptr++;
               }
            } while (*ptr && !isspace(*ptr));

            num = num * sign;

            if (num < -128 || num > 127)
            {
               fprintf(stderr, "Weight %d out of range row %c\n", aminoAcid, num);
               ++errors;
               num = 0;
            }

            row[i] = (char) num;
         }

         if (i < CUSTOM_ALPHABETH_SIZE)
         {
            fprintf(stderr, "Amino acid row %c incomplete\n", aminoAcid);
            ++errors;
         }
      }
   }

   if (errors)
   {
      fprintf(stderr, "Terminating due to errors in matrix %s\n", file);
   }

   fclose(fp);

   ALPHA_SIZE = CUSTOM_ALPHABETH_SIZE;
   ALPHABETH = CUSTOM_ALPHABETH;
   ALPHABETH_VALUE = CUSTOM_ALPHABETH_VALUE;

   printMatrix(matrix);

   return matrix;
}
