/******************************************************************
 Copyright 2006 by Michael Farrar.  All rights reserved.
 This program may not be sold or incorporated into a commercial product,
 in whole or in part, without written consent of Michael Farrar.  For
 further information regarding permission for use or reproduction, please
 contact: Michael Farrar at farrar.michael@gmail.com.
 *******************************************************************/

/**
 Written by Michael Farrar, 2006.
 Modified by Oleksiy Kokhanyy (kohany@gmail.com), 2009 - added backtracking
 matrix in swStripedWord to produce alignment.
 Please send bug reports and/or suggestions to farrar.michael@gmail.com
 and kohany@gmail.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <emmintrin.h>
#include <assert.h>

#include "swsse2.h"
#include "swstriped.h"
#include "score_matrix.h"

const char AMINOACIDS[] = { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'X', 'Y', 'Z', '*' };

const short AMINOACID_VALUE[256] = {
          //                                         *
          //     A   B   C   D   E   F   G   H   I       K   L   M   N
          // P   Q   R   S   T       V   W   X   Y   Z
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 23, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1,  0,  1,  2,  3,  4,  5,  6,  7,  8, -1,  9, 10, 11, 12, -1,
            13, 14, 15, 16, 17, -1, 18, 19, 20, 21, 22, -1, -1, -1, -1, -1,
            -1,  0,  1,  2,  3,  4,  5,  6,  7,  8, -1,  9, 10, 11, 12, -1,
            13, 14, 15, 16, 17, -1, 18, 19, 20, 21, 22, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

const char NUCLEOTIDES[] = { 'A', 'C', 'G', 'K', 'M', 'N', 'R', 'S', 'T', 'W', 'X', 'Y' };
const short NUCLEOTIDES_VALUE[256] = {
   //                                          *
   //      A   B   C   D   E   F   G   H   I       K   L   M   N
   //  P   Q   R   S   T       V   W   X   Y   Z
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1,  0, -1,  1, -1, -1, -1,  2, -1, -1, -1,  3, -1,  4,  5, -1,
      -1, -1,  6,  7,  8, -1, -1,  9, 10, 11, -1, -1, -1, -1, -1, -1,
      -1,  0, -1,  1, -1, -1, -1,  2, -1, -1, -1,  3, -1,  4,  5, -1,
      -1, -1,  6,  7,  8, -1, -1,  9, 10, 11, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

short ALPHA_SIZE = sizeof(NUCLEOTIDES);
const char *ALPHABETH = NUCLEOTIDES;
const short *ALPHABETH_VALUE = NUCLEOTIDES_VALUE;

const signed char xrepMatrix[] = {
      // A   C   G   K   M   N   R   S   T   W   X   Y
         3, -4, -1, -3,  0, -1,  1, -3, -4,  0, -1, -4,
        -4,  3, -4, -3,  0, -1, -4,  0, -2, -3, -1,  1,
        -2, -4,  3,  0, -3, -1,  1,  0, -4, -3, -1, -4,
        -3, -3,  0,  0, -3, -1, -2, -2,  0, -2, -1, -2,
         0,  0, -3, -3,  0, -1, -2, -2, -3, -2, -1, -2,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
         1, -4,  1, -2, -2, -1,  1, -2, -4, -2, -1, -4,
        -3,  0,  0, -2, -2, -1, -2,  0, -3, -3, -1, -2,
        -4, -1, -4,  0, -3, -1, -4, -3,  3,  0, -1,  1,
         0, -3, -3, -2, -2, -1, -2, -3,  0,  0, -1, -2,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -4,  1, -4, -2, -2, -1, -4, -2,  1, -2, -1,  1 };

const signed char *matrix = xrepMatrix;

int gapInit = 10;
int gapExt = 2;

/* ======================================================== */
SwStripedData* swStripedInit_Mat(const unsigned char *querySeq, int queryLength, const signed char *matrix) {
   //    printMatrix(matrix);

   int i, j, k;

   int segSize;
   int nCount;

   int bias;

   //int lenQryByte;
   int lenQryShort;

   int weight;

   short *ps;
   char *pc;

   const signed char *matrixRow;

   size_t aligned;

   SwStripedData *pSwData;

   //lenQryByte = (queryLength + 15) / 16;
   lenQryShort = (queryLength + 7) / 8;

   pSwData = (SwStripedData *) malloc(sizeof(SwStripedData));
   assert(pSwData);

   nCount = 64 + /* slack bytes */
   //         lenQryByte * ALPHA_SIZE + /* query profile byte */
         lenQryShort * ALPHA_SIZE + /* query profile short */
   (lenQryShort * 6); /* vH1, vH2 and vE vEGL*/

   pSwData->pData = (unsigned char *) calloc(nCount, sizeof(__m128i ));
   assert(pSwData->pData);

   /* since we might port this to another platform, lets align the data */
   /* to 16 byte boundries ourselves */
   aligned = ((size_t) pSwData->pData + 15) & ~(0x0f);

   pSwData->pvsQueryProf = (__m128i *) aligned;
   //pSwData->pvsQueryProf = pSwData->pvbQueryProf + lenQryByte * ALPHA_SIZE;

   pSwData->pvH1 = pSwData->pvsQueryProf + lenQryShort * ALPHA_SIZE;
   pSwData->pvH2 = pSwData->pvH1 + lenQryShort;
   pSwData->pvE = pSwData->pvH2 + lenQryShort;
   pSwData->pvEGLLoad = pSwData->pvE + lenQryShort;
   pSwData->pvEGLStore = pSwData->pvEGLLoad + lenQryShort;
   pSwData->pvDIR = pSwData->pvEGLStore + lenQryShort;

   /* Use a scoring profile for the SSE2 implementation, but the layout
    * is a bit strange.  The scoring profile is parallel to the query, but is
    * accessed in a stripped pattern.  The query is divided into equal length
    * segments.  The number of segments is equal to the number of elements
    * processed in the SSE2 register.  For 8-bit calculations, the query will
    * be divided into 16 equal length parts.  If the query is not long enough
    * to fill the last segment, it will be filled with neutral weights.  The
    * first element in the SSE register will hold a value from the first segment,
    * the second element of the SSE register will hold a value from the
    * second segment and so on.  So if the query length is 288, then each
    * segment will have a length of 18.  So the first 16 bytes will  have
    * the following weights: Q1, Q19, Q37, ... Q271; the next 16 bytes will
    * have the following weights: Q2, Q20, Q38, ... Q272; and so on until
    * all parts of all segments have been written.  The last seqment will
    * have the following weights: Q18, Q36, Q54, ... Q288.  This will be
    * done for the entire alphabet.
    */

   /* Find the bias to use in the substitution matrix */
   bias = 127;
   for (i = ALPHA_SIZE * ALPHA_SIZE - 1; i >= 0; i--) {
      if (matrix[i] < bias) {
         bias = matrix[i];
      }
   }
   if (bias > 0) {
      bias = 0;
   }

   /* Fill in the short query profile */
   ps = (short *) pSwData->pvsQueryProf;
   segSize = (queryLength + 7) / 8;
   nCount = segSize * 8;
   for (i = 0; i < ALPHA_SIZE; ++i) {
      matrixRow = matrix + i * ALPHA_SIZE;
      for (j = 0; j < segSize; ++j) {
         for (k = j; k < nCount; k += segSize) {
            if (k >= queryLength) {
               weight = 0;
            } else {
               weight = matrixRow[*(querySeq + k)];
            }
            *ps++ = (unsigned short) weight;
         }
      }
   }

   pSwData->bias = (unsigned short) -bias;

   return pSwData;
}

/* ======================================================== */
SwStripedData* swStripedInit(unsigned char *querySeq, int queryLength) {
   if (matrix == 0) {
      matrix = xrepMatrix;
   }
   return swStripedInit_Mat(querySeq, queryLength, matrix);
}

/* ======================================================== */
SwStripedData* swStripedInitMN(unsigned char *querySeq, int queryLength, signed char match, signed char mismatch) {
   signed char matrix[ALPHA_SIZE * ALPHA_SIZE];
   int i, j;
   memset(matrix, mismatch, (ALPHA_SIZE * ALPHA_SIZE) * sizeof(signed char));
   for (i = 0; i < ALPHA_SIZE; i++) {
      matrix[i * ALPHA_SIZE + i] = match;
   }
   return (SwStripedData *) (swStripedInit_Mat(querySeq, queryLength, matrix));
}

/* ======================================================== */
void swStripedComplete(void *pSwData) {
   SwStripedData *pStripedData = (SwStripedData *) pSwData;

   free(pStripedData->pData);
   free(pStripedData);
}

/* ======================================================== */
void initAlignment(Alignment* aln, int qLen, int dlen) {
   //assert((aln->qAln = (char*) calloc(qLen + dlen, sizeof(char))));
   //assert((aln->dAln = (char*) calloc(qLen + dlen, sizeof(char))));
   aln->qAln = 0;
   aln->dAln = 0;
}

/* ======================================================== */
void deleAlignment(Alignment* aln) {
   if (aln->qAln) {
      free(aln->qAln);
   }
   if (aln->dAln) {
      free(aln->dAln);
   }
}
/**
 * Converts chars to normalized numerical values. Removes
 * illegal characters which can not be converted. Returns new
 * length of the sequence in len pointer.
 */
unsigned char* char2byte_clean(const char* seq, int* len) {
   int alignedLen = (*len) + (17 - (*len) % 16);
   unsigned char *byteSeq = (unsigned char*) malloc(alignedLen + 1);
   assert(byteSeq);
   unsigned short *delChars = 0;
   int widx, ridx;
   for (widx = ridx = 0; ridx < *len; ridx++) {
      if (ALPHABETH_VALUE[seq[ridx]] != -1) {
         byteSeq[widx++] = ALPHABETH_VALUE[seq[ridx]];
      } else {
         char uc = toupper(seq[ridx]);
         if (ALPHABETH == NUCLEOTIDES && (uc == 'B' || uc == 'D' || uc == 'H' || uc == 'V') && ALPHABETH_VALUE['N'] != -1) {
            byteSeq[widx++] = ALPHABETH_VALUE['N'];
         } else {
            if (isgraph(seq[ridx])) {
               if (delChars == 0) {
                  assert((delChars = (unsigned short*) calloc(256, sizeof(unsigned short))));
            }
            delChars[seq[ridx]]++;
         }
      }
   }
}
if (delChars)
{
   unsigned char i;
   for (i = 0; i < 255; i++)
   {
      if (delChars[i] > 0)
      {
         fprintf(stderr, "WARNING: %s:%s:\nWARNING: removed %5d illegal symbols '%c' from input sequence\n", __FILE__, __LINE__, delChars[i], i);
      }
   }
   free(delChars);
}
*len = widx;
alignedLen = (*len) + (17 - (*len) % 16);
for (; widx < alignedLen; widx++)
{
   byteSeq[widx] = ALPHA_SIZE;
}
byteSeq[widx] = '\0';
return byteSeq;
}

/* ======================================================== */
unsigned char* char2byte(const char* seq, int len) {
   int alignedLen = len + (17 - len % 16);
   unsigned char *byteSeq = (unsigned char*) malloc(alignedLen + 1);
   assert(byteSeq);
   int i;
   for (i = 0; i < len; i++) {
      byteSeq[i] = ALPHABETH_VALUE[seq[i]];
   }
   for (; i < alignedLen; i++) {
      byteSeq[i] = ALPHA_SIZE;
   }
   byteSeq[i] = '\0';
   return byteSeq;
}

/* ======================================================== */
char* byte2char(unsigned char* seq, int len) {
   int i;
   for (i = 0; i < len; i++) {
      seq[i] = ALPHABETH[seq[i]];
   }
   seq[i] = '\0';
   return (char*) seq;
}

void loadSWSSE2Matrix(const char* file) {
   if(matrix && matrix != xrepMatrix) {
      free((void*)matrix);
   }
   matrix = readMatrix(file);
}
void setSWSSE2Matrix(const signed char *mat, const short *char2index, const char *index2char, short alphabethSize) {
   if(matrix && matrix != xrepMatrix) {
      free((void*)matrix);
   }
   matrix = (signed char*)malloc(alphabethSize*alphabethSize);
   ALPHA_SIZE = alphabethSize;
   memcpy((void*)matrix, (void*)mat, ALPHA_SIZE*ALPHA_SIZE);
   memcpy((void*)ALPHABETH, (void*)index2char, ALPHA_SIZE);
   memcpy((void*)ALPHABETH_VALUE, (void*)char2index, ALPHA_SIZE);
   printMatrix(matrix);
}
/**
 * qSeq and dSeq are sequences converted with char2byte function
 */
int swStripedAlign_Profile(const unsigned char* qSeq, int queryLength, const unsigned char* dSeq, int dbLength, Alignment* aln, SwStripedData* stripedData) {
   int score = swStripedWord(qSeq, queryLength, dSeq, dbLength, abs(gapInit), abs(gapExt), stripedData->pvsQueryProf, stripedData->pvH1, stripedData->pvH2, stripedData->pvE, stripedData->pvEGLLoad,
         stripedData->pvEGLStore, stripedData->pvDIR, aln);
   return score;
}

/* ======================================================== */
int swStripedAlign(const char* querySeq, int queryLength, const char* dbSeq, int dbLength, Alignment* aln) {
   int ql = queryLength;
   int dl = dbLength;
   unsigned char* qSeq = char2byte_clean(querySeq, &ql);
   unsigned char* dSeq = char2byte_clean(dbSeq, &dl);
   if (ql != queryLength) {
      fprintf(stderr, "WARNING: removed illegal symbols from query sequence.\n");
   }
   if (dl != dbLength) {
      fprintf(stderr, "WARNING: removed illegal symbols from database sequence.\n");
   }
   SwStripedData *stripedData = (SwStripedData *) swStripedInit(qSeq, ql);
   int score = swStripedAlign_Profile(qSeq, ql, dSeq, dl, aln, stripedData);
   free(qSeq);
   free(dSeq);
   swStripedComplete(stripedData);
   return score;
}

/* ======================================================== */
int swStripedWord(const unsigned char *querySeq, int queryLength, const unsigned char *dbSeq, int dbLength, unsigned short gapOpen, unsigned short gapExtend, __m128i *pvQueryProf, __m128i *pvHLoad,
      __m128i *pvHStore, __m128i *pvE, __m128i *pvEGLLoad, __m128i *pvEGLStore, __m128i *pvDIR, Alignment *aln) {
   int i, j;
   int score;

   int cmp;
   int iter = (queryLength + 7) / 8;

   __m128i *pv;

   __m128i vE, vF, vH;

   __m128i vMaxScore;
   __m128i vGapOpen;
   __m128i vGapExtend;
   __m128i vZERO;

   __m128i vMin;
   __m128i vMinimums;
   __m128i vTemp;

   __m128i *pvScore;

   /* Load gap opening penalty to all elements of a constant */
   vGapOpen = _mm_insert_epi16(vGapOpen, gapOpen, 0);
   vGapOpen = _mm_shufflelo_epi16(vGapOpen, 0);
   vGapOpen = _mm_shuffle_epi32(vGapOpen, 0);

   /* Load gap extension penalty to all elements of a constant */
   vGapExtend = _mm_insert_epi16(vGapExtend, gapExtend, 0);
   vGapExtend = _mm_shufflelo_epi16(vGapExtend, 0);
   vGapExtend = _mm_shuffle_epi32(vGapExtend, 0);

   /*  load vMaxScore with the zeros.  since we are using signed */
   /*  math, we will bias the maxscore to -32768 so we have the */
   /*  full range of the short. */
   vMaxScore = _mm_cmpeq_epi16(vMaxScore, vMaxScore);
   vMaxScore = _mm_slli_epi16(vMaxScore, 15);

   vMinimums = _mm_shuffle_epi32(vMaxScore, 0);

   vMin = _mm_shuffle_epi32(vMaxScore, 0);
   vMin = _mm_srli_si128(vMin, 14);

   vZERO = _mm_cmpgt_epi16(vZERO, vZERO);
   /* Zero out the storage vector */
   for (i = 0; i < iter; i++) {
      _mm_store_si128(pvE + i, vMaxScore);
      _mm_store_si128(pvHStore + i, vMaxScore);
      _mm_store_si128(pvEGLStore + i, vZERO);
      _mm_store_si128(pvDIR + i, vZERO);
   }

   short *pathMatrix = (short*) calloc((dbLength + 1) * (queryLength + 1), sizeof(short));
   assert(pathMatrix);

   __m128i vMSK, vMSK2, vEGL, vFGL, vDIR;
   __m128i vONE = _mm_set1_epi16(1);

   //int iMaxScore_i = 0, iMaxScore_j = 0, iMaxScore_r = 0;
   __m128i vMaxScore_i, vMaxScore_j;
   vMaxScore_j = vMaxScore_i = _mm_set1_epi16(0);
   short iMaxScore = -0x8000;

   for (i = 0; i < dbLength; ++i) {
      /* fetch first data asap. */
      pvScore = pvQueryProf + dbSeq[i] * iter;

      /* zero out F. */
      vF = _mm_cmpeq_epi16(vF, vF);
      vF = _mm_slli_epi16(vF, 15);
      /* zero out db gap lengths*/
      vFGL = _mm_cmpgt_epi16(vFGL, vFGL);

      /* load the next h value */
      vH = _mm_load_si128(pvHStore + iter - 1);
      vH = _mm_slli_si128(vH, 2);
      vH = _mm_or_si128(vH, vMin);

      pv = pvHLoad;
      pvHLoad = pvHStore;
      pvHStore = pv;

      pv = pvEGLLoad;
      pvEGLLoad = pvEGLStore;
      pvEGLStore = pv;

      for (j = 0; j < iter; j++) {
         /* load values of vE from previous row (one unit up) */
         vE = _mm_load_si128(pvE + j);

         /* load query gap lengths */
         vEGL = _mm_load_si128(pvEGLLoad + j);

         /* add score to vH */
         vH = _mm_adds_epi16(vH, *pvScore++);

         /**
          *  All code related to backtracing was added by Oleksiy Kohany 02/24/2009
          *  vEGL - query gap lengths
          *  vFGL - database gap lengths
          */
         /* Let's register the position of a new maximum score */
         vMSK = _mm_cmpgt_epi16(vH, vMaxScore);
         vMSK2 = _mm_set1_epi16(i - SHORT_BIAS);
         vMSK2 = _mm_and_si128(vMSK, vMSK2);
         vMaxScore_i = _mm_andnot_si128(vMSK, vMaxScore_i);
         vMaxScore_i = _mm_or_si128(vMaxScore_i, vMSK2);

         vMSK2 = _mm_set1_epi16(j);
         vMSK2 = _mm_and_si128(vMSK, vMSK2);
         vMaxScore_j = _mm_andnot_si128(vMSK, vMaxScore_j);
         vMaxScore_j = _mm_or_si128(vMaxScore_j, vMSK2);

         /* Update highest score encountered this far */
         vMaxScore = _mm_max_epi16(vMaxScore, vH);

         /* adding query and database gaps to traceback vector */
         vMSK = _mm_cmpgt_epi16(vE, vH); //Mask vE>vH
         vDIR = _mm_and_si128(vEGL, vMSK); //COPY vEGL for vE>vH to vDIR
         vMSK = _mm_cmpgt_epi16(vE, vF); //Mask vE>vF
         vDIR = _mm_and_si128(vDIR, vMSK); //Mask out vDIR for vE<=vF
         vMSK2 = _mm_cmpgt_epi16(vF, vH); //Mask vF>vH
         vMSK2 = _mm_and_si128(vFGL, vMSK2); //Copy vFGL for vF>vH
         vMSK2 = _mm_andnot_si128(vMSK, vMSK2); //Remove vFGL where vF<=vE
         vDIR = _mm_subs_epi16(vDIR, vMSK2); //Copy negative vFGL


         /* get max from vH, vE and vF */
         vH = _mm_max_epi16(vH, vE);
         vH = _mm_max_epi16(vH, vF);

         /* adding traceback termination marks to traceback vector */
         vMSK = _mm_cmpeq_epi16(vH, vMinimums);
         vMSK = _mm_and_si128(vMSK, vMinimums);
         vDIR = _mm_or_si128(vDIR, vMSK);

         /* save vH values */
         _mm_store_si128(pvHStore + j, vH);

         /* creating gap initialization scores */
         vH = _mm_subs_epi16(vH, vGapOpen);

         /* creating gap extension scores */
         vE = _mm_subs_epi16(vE, vGapExtend);
         vF = _mm_subs_epi16(vF, vGapExtend);

         /* updating query gap lengths vector */
         vMSK = _mm_cmpgt_epi16(vE, vH);
         vEGL = _mm_and_si128(vEGL, vMSK);
         vEGL = _mm_add_epi16(vEGL, vONE);

         /* updating db gap lengths vector */
         vMSK = _mm_cmpgt_epi16(vF, vH);
         vFGL = _mm_and_si128(vFGL, vMSK);
         vFGL = _mm_add_epi16(vFGL, vONE);

         /* update vE and vF values */
         vE = _mm_max_epi16(vE, vH);
         vF = _mm_max_epi16(vF, vH);

         /* save vE, query gap lengths values, and tracebacks  */
         _mm_store_si128(pvE + j, vE);
         _mm_store_si128(pvEGLStore + j, vEGL);
         _mm_store_si128(pvDIR + j, vDIR);

         /* load the next h value */
         vH = _mm_load_si128(pvHLoad + j);
      }

      /* reset pointers to the start of the saved data */
      j = 0;
      vH = _mm_load_si128(pvHStore + j);

      /*  the computed vF value is for the given column.  since */
      /*  we are at the end, we need to shift the vF value over */
      /*  to the next column. */
      vF = _mm_slli_si128(vF, 2);
      vF = _mm_or_si128(vF, vMin);
      vFGL = _mm_slli_si128(vFGL, 2);
      vTemp = _mm_subs_epi16(vH, vGapOpen);
      vTemp = _mm_cmpgt_epi16(vF, vTemp);
      cmp = _mm_movemask_epi8(vTemp);
      while (cmp != 0x0000) {
         vE = _mm_load_si128(pvE + j);
         vEGL = _mm_load_si128(pvEGLLoad + j);
         vDIR = _mm_load_si128(pvDIR + j);

         vMSK = _mm_cmpgt_epi16(vF, vH);
         vDIR = _mm_and_si128(vFGL, vMSK);
         vMSK = _mm_andnot_si128(vMSK, *(pvDIR + j));
         vDIR = _mm_subs_epi16(vMSK, vDIR);

         vH = _mm_max_epi16(vH, vF);

         /* save vH values */
         _mm_store_si128(pvHStore + j, vH);

         /*  update vE incase the new vH value would change it */
         vH = _mm_subs_epi16(vH, vGapOpen);

         vMSK = _mm_cmpgt_epi16(vE, vH);
         vEGL = _mm_and_si128(vEGL, vMSK);
         vEGL = _mm_add_epi16(vEGL, vONE);

         vE = _mm_max_epi16(vE, vH);

         _mm_store_si128(pvE + j, vE);
         _mm_store_si128(pvEGLStore + j, vEGL);
         _mm_store_si128(pvDIR + j, vDIR);

         /* update vF value */
         vF = _mm_subs_epi16(vF, vGapExtend);
         vFGL = _mm_add_epi16(vFGL, vONE);

         j++;
         if (j >= iter) {
            j = 0;
            vF = _mm_slli_si128(vF, 2);
            vF = _mm_or_si128(vF, vMin);
            vFGL = _mm_slli_si128(vFGL, 2);
         }

         vH = _mm_load_si128(pvHStore + j);

         vTemp = _mm_subs_epi16(vH, vGapOpen);
         vTemp = _mm_cmpgt_epi16(vF, vTemp);
         cmp = _mm_movemask_epi8(vTemp);
      }
      /*
       * START
       * Filling in trace matrix for alignment recovery
       */
      short *path = (short*) pvDIR;
      int values_per_vector = 8;
      int wi, qi;
      int mi = 1;
      //For each value in 128bit vector
      for (wi = 0; wi < values_per_vector; wi++) {
         //For each 8-word iteration
         for (qi = 0; qi < iter; qi++) {
            int queryPos = wi + qi * values_per_vector;
            if (mi <= queryLength) {
               int index = mi * (dbLength + 1) + i + 1;
               pathMatrix[index] = path[queryPos];
               mi++;
            }
         }
      }
      /*
       * END
       */
   }

   short *p_iMaxScore_i = (short*) &vMaxScore_i;
   short *p_iMaxScore_j = (short*) &vMaxScore_j;
   short *p_iMaxScore = (short*) &vMaxScore;
   int iMaxScore_i = 0, iMaxScore_j = 0, iMaxScore_r = 0;
   for (i = 0; i < 8; i++) {
      if (iMaxScore < p_iMaxScore[i]) {
         iMaxScore = p_iMaxScore[i];
         iMaxScore_i = p_iMaxScore_i[i];
         iMaxScore_j = p_iMaxScore_j[i];
         iMaxScore_r = i;
      }
   }
   iMaxScore_i += SHORT_BIAS;
   int iMaxX = iMaxScore_i + 1;
   int iMaxY = iMaxScore_r * iter + iMaxScore_j + 1;
   aln->score = (short) iMaxScore + SHORT_BIAS;
   if (aln->score) {
      traceBack_Trace_Word(querySeq, queryLength, dbSeq, dbLength, pathMatrix, iMaxX, iMaxY, aln);
   }
   free(pathMatrix);
   return aln->score;
}

/* ======================================================== */
void printAlignmentMatrix(short* scoreMat, int height, int width, char* rowHead, char* colHead) {
   fprintf(stderr, "\n");
   fprintf(stderr, ">Q\n%s\n", rowHead);
   fprintf(stderr, ">D\n%s\n", colHead);

   fprintf(stderr, "%5c%5c", ' ', ' ');
   int x, y;
   for (x = 0; x < width - 1; x++) {
      fprintf(stderr, "%c:%-3d", colHead[x], x + 1);
   }
   fprintf(stderr, "\n");
   for (y = 0; y < height; y++) {
      if (y > 0) {
         fprintf(stderr, "%c:%-3d", rowHead[y - 1], y);
      } else {
         fprintf(stderr, "%5c", ' ');
      }
      for (x = 0; x < width; x++) {

         fprintf(stderr, "%-5d", scoreMat[y * width + x]);
      }
      fprintf(stderr, "\n");
   }
}

/* ======================================================== */
void printTraceMatrix(short* scoreMat, int height, int width, char* rowHead, char* colHead) {
   fprintf(stderr, "\n");
   fprintf(stderr, ">Q\n%s\n", rowHead);
   fprintf(stderr, ">D\n%s\n", colHead);

   fprintf(stderr, "%5c%5c", ' ', ' ');
   int x, y;
   for (x = 0; x < width - 1; x++) {
      fprintf(stderr, "%c:%-3d", colHead[x], x + 1);
   }
   fprintf(stderr, "\n");
   for (y = 0; y < height; y++) {
      if (y > 0) {
         fprintf(stderr, "%c:%-3d", rowHead[y - 1], y);
      } else {
         fprintf(stderr, "%5c", ' ');
      }
      for (x = 0; x < width; x++) {
         switch (scoreMat[y * width + x]) {
         case 0:
            fprintf(stderr, "%-5c", '\\');
            break;
         case -0x8000:
            fprintf(stderr, "%-5c", 'X');
            break;
         default:
            fprintf(stderr, "%-5d", scoreMat[y * width + x]);
         }
      }
      fprintf(stderr, "\n");
   }
}

/**
 * Traces back using Gotoh's trace back matrix.
 * iMaxX, iMaxY - position of maximum score corresponding to traceback matrix.
 * aln - pointer to Alignment object to save alignment.
 * local - if 0, then unaligned flanks
 */

int traceBack_Trace_Word(const unsigned char* querySeq, int queryLength, const unsigned char* dbSeq, int dbLength, short* pathMat, int iMaxX, int iMaxY, Alignment *aln) {
   int i, j;
   int matHeight = queryLength + 1;
   int matWidth = dbLength + 1;

   char *qSeq = (char*) malloc(matHeight * sizeof(char));
   char *dSeq = (char*) malloc(matWidth * sizeof(char));
   char *qAln = (char*) malloc(matHeight * matWidth * sizeof(char));
   char *dAln = (char*) malloc(matHeight * matWidth * sizeof(char));
   assert(qSeq);
   assert(dSeq);
   assert(qAln);
   assert(dAln);
   qSeq[matHeight - 1] = '\0';
   dSeq[matWidth - 1] = '\0';
   for (i = matHeight - 2; i >= 0; i--) {
      qSeq[i] = ALPHABETH[(char) querySeq[i]];
   }
   for (i = matWidth - 2; i >= 0; i--) {
      dSeq[i] = ALPHABETH[(char) dbSeq[i]];
   }

#ifdef DEBUG
   printTraceMatrix(pathMat, matHeight, matWidth, qSeq, dSeq);
#endif

   aln->qEnd = matHeight - 2;
   aln->dEnd = matWidth - 2;
   aln->qStart = 0;
   aln->dStart = 0;
   int qi = 0, di = 0;

   aln->qEnd = iMaxY;
   aln->dEnd = iMaxX;
   int x = iMaxX;
   int y = iMaxY;

#ifdef DEBUG

   fprintf(stderr, "START [%d:%d]\n", x, y);
#endif

   do {
      int i = y * matWidth + x;
      if (pathMat[i] == -0x8000) {
         break;
      }
      if (pathMat[i] == 0) {
         qAln[qi++] = qSeq[--y];
         dAln[di++] = dSeq[--x];
      } else {
         if (pathMat[i] > 0) {
            int qugl = pathMat[i];
            while (qugl-- > 0) {
               qAln[qi++] = '-';
               dAln[di++] = dSeq[--x];
            }
         } else {
            if (pathMat[i] < 0) {
               int dbgl = -pathMat[i];
               while (dbgl-- > 0) {
                  qAln[qi++] = qSeq[--y];
                  dAln[di++] = '-';
               }
            }
         }
      }
   } while (x > 0 && y > 0);

   aln->qStart = y + 1;
   aln->dStart = x + 1;
   dAln[di] = 0;
   qAln[qi] = 0;

#ifdef DEBUG

   fprintf(stderr, "%dx%d\n", di, qi);
   fprintf(stderr, ">db\n%s\n", dAln);
   fprintf(stderr, ">qu\n%s\n", qAln);
#endif

   aln->alnLen = qi;
   if (aln->qAln) {
      free(aln->qAln);
   }
   if (aln->dAln) {
      free(aln->dAln);
   }
   aln->qAln = calloc(aln->alnLen + 1, sizeof(char));
   aln->dAln = calloc(aln->alnLen + 1, sizeof(char));
   for (i = 0, qi--; qi >= 0; qi--, i++) {
      //printf("q[%d]=%d\n", qi, qAln[qi]);
      aln->qAln[i] = qAln[qi];
      aln->dAln[i] = dAln[qi];
   }
   aln->qAln[i] = '\0';
   aln->dAln[i] = '\0';
   free(qAln);
   free(dAln);
   free(qSeq);
   free(dSeq);
   return di;
}

/**
 * Traces back through Smith Waterman alignment score matrix.
 */
int traceBack_Score_Word(unsigned char* querySeq, int queryLength, unsigned char* dbSeq, int dbLength, unsigned short* scoreMat, int iGI, int iGX, Alignment *aln, short local) {
   int i, j;
   int matHeight = queryLength + 1;
   int matWidth = dbLength + 1;

   char qSeq[matHeight];
   char dSeq[matWidth];
   char qAln[matHeight * matWidth];
   char dAln[matHeight * matWidth];
   qSeq[matHeight - 1] = '\0';
   dSeq[matWidth - 1] = '\0';
   for (i = matHeight - 2; i >= 0; i--) {
      qSeq[i] = ALPHABETH[(char) querySeq[i]];
   }
   for (i = matWidth - 2; i >= 0; i--) {
      dSeq[i] = ALPHABETH[(char) dbSeq[i]];
   }

#ifdef DEBUG
   printAlignmentMatrix(scoreMat, matHeight, matWidth, qSeq, dSeq);
#endif

   int x = matWidth - 1, y = matHeight - 1;
   aln->qEnd = matHeight - 2;
   aln->dEnd = matWidth - 2;
   aln->qStart = 0;
   aln->dStart = 0;
   int qi = 0, di = 0;

   if (local) {
      int maxScorePos = matHeight * matWidth - 1;
      int maxScore = scoreMat[maxScorePos];
      for (i = maxScorePos; i >= 0; i--) {
         if (maxScore < scoreMat[i]) {
            maxScorePos = i;
            maxScore = scoreMat[i];
         }
      }

      x = maxScorePos % matWidth;
      y = maxScorePos / matWidth;
      aln->qEnd = y;
      aln->dEnd = x;
#ifdef DEBUG

      fprintf(stderr, "matHeight=%d, matWidth=%d, maxScore=%d, maxScorePos=%d, maxX=%d, maxY=%d\n", matHeight, matWidth, maxScore, maxScorePos, x, y);
#endif

   }

#ifdef DEBUG

   fprintf(stderr, "START [%d:%d]\n", x, y);
#endif

   int qGapLength = 0;
   int dGapLength = 0;
   do {
      int i = y * matWidth + x;
      int ni = (y - 1) * matWidth + x;
      int wi = y * matWidth + x - 1;
      int nwi = (y - 1) * matWidth + x - 1;

#ifdef DEBUG

      fprintf(stderr, "C=%d\tW=%d\tNW=%d\tN=%d\t", scoreMat[i], scoreMat[wi], scoreMat[nwi], scoreMat[ni]);
#endif

      if (local && scoreMat[nwi] <= 0) {
#ifdef DEBUG
         fprintf(stderr, "Encountered 0 at [%d:%d]\n", x-1, y-1);
#endif

         qAln[qi++] = qSeq[--y];
         dAln[di++] = dSeq[--x];
         break;
      }
      if (scoreMat[nwi] >= scoreMat[wi] && scoreMat[nwi] >= scoreMat[ni]) {
         qGapLength = 0;
         dGapLength = 0;
         qAln[qi++] = qSeq[--y];
         dAln[di++] = dSeq[--x];
#ifdef DEBUG

         fprintf(stderr, "NW [%d:%d]\n", x, y);
#endif

      } else {
         if (scoreMat[wi] >= scoreMat[ni]) {
            qGapLength++;
            qAln[qi++] = '-';
            dAln[di++] = dSeq[--x];
#ifdef DEBUG

            fprintf(stderr, "W [%d:%d]\n", x, y);
#endif

         } else {
            dGapLength++;
            qAln[qi++] = qSeq[--y];
            dAln[di++] = '-';
#ifdef DEBUG

            fprintf(stderr, "N [%d:%d]\n", x, y);
#endif

         }
      }
   } while (x > 0 && y > 0);

   if (local) {
      aln->qStart = y + 1;
      aln->dStart = x + 1;
   } else {
      while (y > 1) {
         qAln[qi++] = qSeq[--y];
         dAln[di++] = '-';
      }
      while (x > 1) {
         qAln[qi++] = '-';
         dAln[di++] = dSeq[--x];
      }
   }
   dAln[di] = 0;
   qAln[qi] = 0;

#ifdef DEBUG

   fprintf(stderr, "%dx%d\n", di, qi);
   fprintf(stderr, ">db\n%s\n", dAln);
   fprintf(stderr, ">qu\n%s\n", qAln);
#endif

   aln->alnLen = qi;
   for (i = 0, qi--; qi >= 0; qi--, i++) {
      //printf("q[%d]=%d\n", qi, qAln[qi]);
      aln->qAln[i] = qAln[qi];
      aln->dAln[i] = dAln[qi];
   }
   aln->qAln[i] = '\0';
   aln->dAln[i] = '\0';

   return di;
}

